/*  CommandIR Library
    CommandIR III
    Feb 2011, Matthew Bodkin
 */

#include "libcmdir.h"

usb_dev_handle *cmdir_udev = 0;
char incoming_packet[1024];
char * rec_filename = NULL;
char * send_filename = NULL;
unsigned int tx_bitmask = 0;
char outgoingBuffer[4096];


int local_buffer[1000];
int write_local_buffer = 0;
int currentPCA_Frequency = (48000000/38000);	

// Receive a standard string on the Command Line and transmit it
// eg. 0000 006c 0022 0002 0156 00ac 0015 0016 0015 0041 0015 0016 0015 0016 0015 0041
// 0000 + Freq + # + rpt

char xtod(char c)
{
  if (c>='0' && c<='9') return c-'0';
  if (c>='A' && c<='F') return c-'A'+10;
  if (c>='a' && c<='f') return c-'a'+10;
  return c=0;
}


int convert_quad(char * a, int len)
{
  int y, d, n, p;
  p = 1;
  n = 0;
  for(y = len - 1 ; y >= 0; y--)
  {
    d = xtod(a[y]) * p;
    n += d;
    p *= 0x10;
  }
  return n;
}

int hardware_scan()
{
  // Find the first CommandIR device:
  int r;
	struct usb_bus *bus;
	struct usb_device *dev;

  int read_retval;

  usb_init(); 
  
	if(!usb_find_busses()) {
	  printf("No busses found!\n");
	  return 0;
	}
	
	if(!usb_find_devices()) {
	  printf("No devices found!\n");
		return 0;
	}
	
	if(cmdir_udev) return 0;
		
  for (bus = usb_busses; bus; bus = bus->next)
	{
		for (dev = bus->devices; dev; dev = dev->next)
		{
			if ( (dev->descriptor.idVendor == USB_CMDIR_VENDOR_ID) && 
				(dev->descriptor.idProduct == 3) )
			{
			  cmdir_udev = usb_open(dev);
        r = usb_claim_interface(cmdir_udev, 0);
        if(r>=0) 
        {
			    return 1;
			  } else {
			    // Is this a permission problem, or is another commandir util running?
			    printf(" Unable to claim CommandIR (%d).\n", r);
			    exit(1);
			  }
	    }
	   }
	}
	
	printf("No CommandIR found\n");
	return 0;
}

struct sendir * tx_char_2_struct(char * buf)
{
  struct commandir_3_tx_signal * ptx;
  static struct sendir retSend;
  float freq;
  unsigned char * pWrite, pulse_toggle = 1;
  unsigned char * freqArg;
  unsigned char * importArg;
  int x, tx_count;
  unsigned int n;
  int outgoingLen;
  
  outgoingLen = ((strlen(buf) / 5) - 4) * 2 + sizeof(struct commandir_3_tx_signal);
  
/*  printf("The leng(buf)=%d, so the estimated USB is: %d\n", 
    strlen(buf), 
    outgoingLen );
 */
  
  retSend.buffer = malloc(outgoingLen + 1);
  
  freqArg = &buf[5];
  importArg = &buf[20];
  
  pWrite = &retSend.buffer[ sizeof(struct commandir_3_tx_signal) ];
  ptx = (struct commandir_3_tx_signal *)retSend.buffer;
  ptx->tx_bit_mask1 = tx_bitmask ;
  ptx->tx_bit_mask2 = tx_bitmask >> 16;
  ptx->tx_min_gap = 0xffff;
  ptx->tx_signal_count = 0;
  
  
  freq = (float)1000000/((float)convert_quad(freqArg, 4)*0.241246);
  ptx->pulse_width = 48000000/freq;
  ptx->pwm_offset = ptx->pulse_width / 2;
  
  tx_count = sizeof(struct commandir_3_tx_signal);
  
  retSend.byteCount = sizeof(struct commandir_3_tx_signal);
  
  printf("Transmitting '%s' at%5.1fkHz\n", send_filename, freq/1000.0);
  
  while(*importArg != 0)
  {
    n = convert_quad(importArg, 4);
    if(n == 0) break;
    *pWrite++ = n & 0xff;  // low byte
    *pWrite++ = (n >> 8) | (pulse_toggle << 7);
    retSend.byteCount += 2;
    ptx->tx_signal_count++;
//    printf("Number is %d (pulse? %d)\n", n, pulse_toggle);
    pulse_toggle = !pulse_toggle;
    tx_count+=2;
    importArg += 5;
  }


  return &retSend;
}

char * load_file(char * filename)
{
  FILE *in;
  int filesize;
  char * fileBuffer;
  
  in = fopen(filename, "rt");
  if(!in)
  {
    printf("File not found: %s\n", filename);
    return NULL;
  }
  fseek(in, 0, SEEK_END);
  filesize = ftell(in);
  fseek(in, 0, SEEK_SET);
  
  if(filesize > 65500)
  {
    printf("File too big\n");
    return NULL;
  }
  
  fileBuffer = malloc(filesize + 1);
  
  fgets(fileBuffer, filesize, in);

  fclose(in);

  return fileBuffer;
}




int commandir_sendir(struct sendir * tx)
{
  return commandir_send(tx->buffer, tx->byteCount);
}

int commandir_send(char * packet, int bytes)
{
  int send_status, x;
  char * curPacket;

  if(!cmdir_udev && !hardware_scan())
    return 0;
  
  curPacket = packet;
  
  while(bytes > 0)
  {
  
	  send_status=usb_bulk_write(
		  cmdir_udev,
		  2,
		  curPacket,
		  bytes > 64 ? 64 : bytes,
		  50);
		if(bytes > 64)
		{
		  curPacket+=64;
		  bytes-=64;
		} else {
		  bytes = 0;
		}
//    printf(" send_status: %d\n", send_status);
  }
  return send_status;
}


int kbhit (void)
{
  struct timeval tv;
  fd_set rdfs;

  tv.tv_sec = 0;
  tv.tv_usec = 0;

  FD_ZERO(&rdfs);
  FD_SET (STDIN_FILENO, &rdfs);

  select(STDIN_FILENO+1, &rdfs, NULL, NULL, &tv);
  return FD_ISSET(STDIN_FILENO, &rdfs);

}

int commandir_rec()
{
  //  Let's do a loop and print the recorded signals in HEX format
  int read_retval;
  int read_size = 64;
  int read_timeout = 5000;
  int read_endpoint = 0x83;

  printf("Press the key a file times on your remote.  \nPress 'enter' on the keyboard to select the latest IR code.\n");
  
  while(!kbhit())
  {
    read_retval = usb_bulk_read(
      cmdir_udev,
      0x83,
      incoming_packet,
      64,
      5000);

    if(read_retval > 0)
    {
      commandir3_convert_RX(incoming_packet, read_retval -1);
    }
  }
  printf("Accepted.\n", getchar());
  
  return 0;

}


int lirc_pipe_write(int * readyDataByte)
{
  static niceprint = 1;
  return 0;
  if(++niceprint > 6) 
  {
    niceprint = 1;
    printf("\n");
  }
  
  // So now we want to build items recognizable as specific codes; split by >50,000
  
  printf("%d  ", (0x7ff) & *readyDataByte);
  
  if(*readyDataByte > 50000)
  {
    printf("\n\n");
    niceprint = 0;
  }
  
  return 0;

}





void add_pulses(int pcount)
{
  local_buffer[write_local_buffer++] = pcount;
}

void add_space(int slen)
{
  local_buffer[write_local_buffer++] = slen;
}

void dump_signal()
{
  int x;
  FILE * pFile;
  int freq;
  int fileFreq;

  // To file:
  pFile = fopen (rec_filename == NULL ? "testrec.txt" : rec_filename,"w");
  if(pFile==NULL)
  {
    printf("Error opening output file.\n");
    exit(0);
  }
  
  freq = 1000000/(currentPCA_Frequency/48);
  fileFreq = (1000000/ (freq * 0.241246) );

  printf("At %5.1fkHz: ", freq/1000.0 );

  printf("0000 %04X %04X %04X ", fileFreq, write_local_buffer, 0);
  fprintf(pFile, "0000 %04X %04X %04X ", fileFreq, write_local_buffer, 0);
  
  for(x = 0; x < write_local_buffer; x++)
  {
    printf("%04X ", (x % 2 == 0) ? local_buffer[x] : local_buffer[x] * 48 / currentPCA_Frequency );
    fprintf(pFile, "%04X ", (x % 2 == 0) ? local_buffer[x] : local_buffer[x] * 48 / currentPCA_Frequency );
  }
  printf("\nPress 'Enter' to accept this IR code for %s.\n", rec_filename);
  fprintf(pFile, "\n");
  
  fclose (pFile);
  write_local_buffer = 0;
}

// Originally written in LIRC's hw_commandir.c
int commandir3_convert_RX(unsigned char *rxBuffer,
	int numNewValues)
{
	static unsigned char incomingBuffer[MAX_INCOMING_BUFFER + 30];
	static unsigned char switchByte = 0;
	static int incomingBuffer_Write = 0, incomingBuffer_Read = 0;
	static unsigned int	buffer_write;
	static float currentPWM = 0;
	
	int i, mySize, read_num;
	int currentSize, expectingBytes;
	int packet_number;
	static int mcu_rx_top_location;
	
	struct usb_rx_pulse3 * a_usb_rx_pulse;
	struct usb_rx_space3 * a_usb_rx_space;
	struct usb_rx_pulse_def3 * a_usb_rx_pulse_def;
	struct usb_rx_demod_pulse * a_usb_rx_demod_pulse;
	
	if(numNewValues > 0)
	{
		packet_number = rxBuffer[0];
		expectingBytes = rxBuffer[1] + rxBuffer[2] * 256;
		if(numNewValues != (expectingBytes + 5))
			printf("MCU top is now: %d (a change of: %d)\n", 
				rxBuffer[3] + rxBuffer[4] * 256, (rxBuffer[3] + rxBuffer[4] * 256) - 
				mcu_rx_top_location);
		mcu_rx_top_location = rxBuffer[3] + rxBuffer[4] * 256;
		
		if(numNewValues != (expectingBytes + 5))
			printf(
				"USB received: %d, expectingBytes: %d. Hex data (headers: %d %d %d):\t"
				, numNewValues, expectingBytes, rxBuffer[0], rxBuffer[1], rxBuffer[2]);
		for (i = 5; i < expectingBytes + 5; i++)
		{
			if(numNewValues != (expectingBytes + 5)) 	
				printf("%02x@%d\t", rxBuffer[i], incomingBuffer_Write);
 
			incomingBuffer[incomingBuffer_Write++] = rxBuffer[i];
			if(incomingBuffer_Write >= MAX_INCOMING_BUFFER)
				incomingBuffer_Write = 0;
		}
		
		if(incomingBuffer_Write > incomingBuffer_Read)
		{
			currentSize = incomingBuffer_Write - incomingBuffer_Read;
		} else {
			currentSize = (MAX_INCOMING_BUFFER - incomingBuffer_Read) + 
				incomingBuffer_Write;
		}
		
		while(currentSize > 0)
		{
			// Make sure the entire struct is received before trying to use
			switch(incomingBuffer[incomingBuffer_Read])
			{
				case USB_RX_PULSE:
					mySize = 2;
					break;
				case USB_RX_SPACE:
					mySize = 4;
					break;
				case USB_RX_PULSE_DEMOD:
					mySize = 4;
					break;
				case USB_RX_SPACE_DEMOD:
					mySize = 4;
					break;
				case USB_RX_PULSE_DEF:
					mySize = 4;
					break;
				case USB_NO_DATA_BYTE:
//					read_num = write(child_pipe_write, lirc_zero_buffer, 
//						sizeof(lirc_t)*insert_fast_zeros);
					mySize = 0;
					break;
				default:
					printf("Unknown struct identifier: %02x at %d\n", 
						incomingBuffer[incomingBuffer_Read], incomingBuffer_Read);
					mySize = 0;
					break;
			}
			
			if(currentSize >= (mySize+1))	
			{
				if(incomingBuffer_Read + mySize >= MAX_INCOMING_BUFFER)
				{
					memcpy(&incomingBuffer[MAX_INCOMING_BUFFER], &incomingBuffer[0], 
						mySize + 1);	// +1 to compensate for header
				}
				
				switchByte = incomingBuffer[incomingBuffer_Read++];
				if(incomingBuffer_Read >= MAX_INCOMING_BUFFER)
					incomingBuffer_Read -= MAX_INCOMING_BUFFER;
				currentSize--;

				switch(switchByte)
				{
				case USB_NO_DATA_BYTE:
					// Hold this feature for now
					//	buffer_write = 0;
					//	lirc_pipe_write(&buffer_write);
					break;
					
				case USB_RX_PULSE:
					a_usb_rx_pulse = (struct usb_rx_pulse3 *)
						&incomingBuffer[incomingBuffer_Read];
/*					buffer_write =  a_usb_rx_pulse->t0_count * 
						(1000000/ (1/( ((float)(currentPCA_Frequency))/48000000) ) ) ;
					buffer_write |= PULSE_BIT; 
					lirc_pipe_write(&buffer_write); */
					add_pulses(a_usb_rx_pulse->t0_count);
					break;
					
				case USB_RX_SPACE:
					a_usb_rx_space = (struct usb_rx_space3 *)
						&incomingBuffer[incomingBuffer_Read];
					buffer_write = (a_usb_rx_space->pca_offset + 
						(a_usb_rx_space->pca_overflow_count) * 0xffff)/48;
					lirc_pipe_write(&buffer_write);
					if(buffer_write < 50000)
					  add_space(buffer_write);
					else
					  dump_signal();
					break;
					
				case USB_RX_PULSE_DEF:
					a_usb_rx_pulse_def = (struct usb_rx_pulse_def3 *)
						&incomingBuffer[incomingBuffer_Read];
					currentPWM = ((float)(a_usb_rx_pulse_def->pwm))/48;

/*					printf(" Pulse Def, modulation frequency: %f us, pwm: %fus; Duty Cycle: %f%%\n",
							1/( ((float)(a_usb_rx_pulse_def->frequency))/48000000),
							currentPWM,
							100* ((((float)(a_usb_rx_pulse_def->pwm))/48) / 
								(((float)(a_usb_rx_pulse_def->frequency)/48)) ));
								*/
					currentPCA_Frequency = a_usb_rx_pulse_def->frequency;
					break;
					
				case USB_RX_PULSE_DEMOD:
					a_usb_rx_demod_pulse = (struct usb_rx_demod_pulse*)
						&incomingBuffer[incomingBuffer_Read];
					buffer_write = (a_usb_rx_demod_pulse->pca_offset + 
						(a_usb_rx_demod_pulse->pca_overflow_count) * 0xffff)/48;
					buffer_write |= PULSE_BIT;
					lirc_pipe_write(&buffer_write);
					break;
					
				case USB_RX_SPACE_DEMOD:
					a_usb_rx_demod_pulse = (struct usb_rx_demod_pulse*)
						&incomingBuffer[incomingBuffer_Read];
					buffer_write = (a_usb_rx_demod_pulse->pca_offset + 
						(a_usb_rx_demod_pulse->pca_overflow_count) * 0xffff)/48;
					lirc_pipe_write(&buffer_write);
					break;
				}
				
				incomingBuffer_Read +=mySize;
				if(incomingBuffer_Read >= MAX_INCOMING_BUFFER)
					incomingBuffer_Read -= MAX_INCOMING_BUFFER;
				currentSize -= mySize;
			} else {
				break; // from while
			}
		}
	}
	return 0;
}



